#AtomsFiller.py

import sys
sys.path.append("..")
from MDWebService import *
from AtomsDecorator import *


class RectangleDecorator(AtomsDecorator):

    def __init__(self,Parameters,ioControl,MD,model):
        AtomsDecorator.__init__(self,Parameters,ioControl,MD,model)
        self.numConstrainedLeaves = 0 # needed by base class!

        
    def MakeLeaves(self):

        constrZoneWidth = self.potential.GetCutoffDistance()
        # declare parallelepiped cluster initializer
        self.boundaryVectors = [(self.Parameters["length"]+constrZoneWidth)*array(self.Parameters["rectangleAxes"][0]),
                                (self.Parameters["width"]+constrZoneWidth)*array(self.Parameters["rectangleAxes"][1]),
                                (self.Parameters["height"]+constrZoneWidth)*array(self.Parameters["rectangleAxes"][2])]
        # must transpose because C++ indexes differently from python
        transposedBoundaryVectors = array(Transpose(self.boundaryVectors))
        parallelepiped = self.MD.ParallelepipedClusterInitializer(transposedBoundaryVectors, self.lattice)
        if self.Parameters["parallel"]:
            parappd = self.MD.ParallelAtomsInitializer(parallelepiped)
            initializer = parappd
        else:
            initializer = parallelepiped


        # for each material (or grain)
        LOAId = 0
        for matID in self.model.GetMaterialIds():

            # get rotation
            rotation = array(self.model.GetMaterialData(matID))

            # get list of elems with current material id
            elems = []
            for elemID in self.model.GetElemIds():
                if self.model.GetElemData(elemID)[1] == matID:
                    elems.append(elemID)

            self.ioControl.Print(4,"Rotation = " + repr(rotation))

            # generate rectangular cluster of atoms
            parallelepiped.SetCenter(self.Parameters["center"])
            self.lattice.ResetOrientation()
            self.lattice.RotateLatticeVectors(rotation)
            initializer.Create(self.BulkLOAs[LOAId])

            # remove atoms outside material region
            self.ioControl.Print(3,"Removing extra atoms...")
            self.RemoveExtraAtoms(elems,self.BulkLOAs[LOAId])

            # increment ListOfAtoms
            LOAId += 1
            

    def MakeConstrainedLeaves(self):

        # make list of faces with surface points and normals
        self.faceInfo = [0,0,0,0,0,0]
        centerVector = self.Parameters["center"] 
        for i in xrange(3):
            boundaryVector = self.boundaryVectors[i] 
            surface1 = centerVector+0.5*boundaryVector
            normal1 = -1.0*self.Parameters["rectangleAxes"][i]
            surface2 = centerVector-0.5*boundaryVector
            normal2 = self.Parameters["rectangleAxes"][i]
            self.faceInfo[i*2] = [normal1,surface1]
            self.faceInfo[i*2+1] = [normal2,surface2]
        self.ioControl.Print(4,self.faceInfo)

        # make constrZoneMap
        self.constrZoneMap = []


        constrZoneWidth = self.potential.GetCutoffDistance()
        self.ioControl.Print(3,"Constrained zone width = " + repr(constrZoneWidth))
        self.ioControl.Print(3,"Finding atoms in constrained zones...")
        # loop over materials
        for matID in xrange(len(self.model.GetMaterialIds())):
            tempConstrZoneMap = []
            removeList = []
            tempConstrLOAs = []
            
            positions = self.bulk.GetBranch(matID).GetCartesianPositions()
            # loop over atoms
            for atomIdx in xrange(self.bulk.GetBranch(matID).GetNumber()):
                position = positions[atomIdx]
                cartesian = array([position[0],position[1],position[2]]) 
                adjacentFaces = []
                # loop over faces 
                for faceID in xrange(6):
                    surfaceNormal = self.faceInfo[faceID][0] 
                    surfacePoint = self.faceInfo[faceID][1]
                    perpDist = abs(dot(surfaceNormal,surfacePoint-cartesian))
                    self.ioControl.Print(4,"Atom " + repr(atomIdx) + " at position " + repr(cartesian) + " is " + repr(perpDist) + " away from face " + repr(faceID) + " with point " + repr(surfacePoint))
                    if perpDist < constrZoneWidth:
                        adjacentFaces.append(faceID)


                if len(adjacentFaces) > 0:
                    if adjacentFaces not in tempConstrZoneMap:
                        tempConstrZoneMap.append(adjacentFaces)
                        LOAId = len(tempConstrLOAs)
                        tempConstrLOAs.append(self.MD.DynamicListOfAtoms(0))
                    else:
                        for k in xrange(len(tempConstrZoneMap)):
                            if adjacentFaces == tempConstrZoneMap[k]: LOAId = k
                    removeList.append(atomIdx)
                    self.ioControl.Print(4,"Moving Atom: " + repr(atomIdx) + " to constrained zone for face(s): " + repr(adjacentFaces))
                    singleAtom = self.MD.DynamicListOfAtoms(1)
                    for coord in xrange(3):
                        singleAtom.SetCartesianPosition(0,coord,cartesian[coord])
                    tempConstrLOAs[LOAId].Merge(singleAtom)

            for faces in tempConstrZoneMap:
                self.constrZoneMap.append(faces)

            self.ConstrLOAs.extend(tempConstrLOAs)

            # remove atoms from bulkLOA
            self.ioControl.Print(3,"Moving " + repr(len(removeList)) + " out of " + repr(self.bulk.GetBranch(matID).GetNumber()) + " atoms to constrained Zones")
            self.bulk.GetBranch(matID).RemoveAtoms(removeList)


        # find which LOAs are in endcaps for cohesive law measurement
        if self.Parameters["deformationMethod"] == "cohesiveLaw":
            self.endcap1 = self.MD.DynamicListOfAtoms(0)
            self.endcap2 = self.MD.DynamicListOfAtoms(0)
            for LOAId in xrange(len(self.constrZoneMap)):
                if 0 in self.constrZoneMap[LOAId]:
                    self.endcap1.Attach(self.ConstrLOAs[LOAId])
                if 1 in self.constrZoneMap[LOAId]:
                    self.endcap2.Attach(self.ConstrLOAs[LOAId])
        
        
        self.ioControl.Print(4,"ConstrZoneMap: " + repr(self.constrZoneMap))
        self.ioControl.Print(3,"length of constrZoneMap: " + repr(len(self.constrZoneMap)))
        self.ioControl.Print(3,"length of ConstrLOAs: " + repr(len(self.ConstrLOAs)))




    def ImplementConstraints(self):
        
        # implement constraints
        self.ioControl.Print(3,"Implementing Constraints...")
        self.constraints = []
        for LOAId in xrange(len(self.constrZoneMap)):
            if len(self.constrZoneMap[LOAId]) == 1 and self.Parameters["bc"] == "rollered":
                normalVect = self.faceInfo[self.constrZoneMap[LOAId][0]][0]
                normal = array([normalVect[0],normalVect[1],normalVect[2]])
                self.constraints.append(self.MD.FixedInPlane(self.ConstrLOAs[LOAId].GetCartesianPositions(),normal))
            elif len(self.constrZoneMap[LOAId]) == 2 and self.Parameters["bc"] == "rollered":
                normalVect1 = self.faceInfo[self.constrZoneMap[LOAId][0]][0]
                normalVect2 = self.faceInfo[self.constrZoneMap[LOAId][1]][0]
                tangentVect = CrossProduct(normalVect1,normalVect2)
                tangent = array([tangentVect[0],tangentVect[1],tangentVect[2]])
                self.constraints.append(self.MD.FixedInLine(self.ConstrLOAs[LOAId].GetCartesianPositions(),tangent))
            elif len(self.constrZoneMap[LOAId]) == 3 or self.Parameters["bc"] == "constrained":
                self.constraints.append(self.MD.FixedBody(self.ConstrLOAs[LOAId].GetCartesianPositions()))
            self.ConstrLOAs[LOAId].SetConstraint(self.constraints[LOAId])
 
                   
